# -*- coding: utf-8 -*-
"""

Summary
-------

Signal processing functions

Resampling  & smoothing.


-------------------------------------------------------------------------------


"""

import math

import numpy as np
import scipy.linalg as LA
from numpy import abs, isnan, linspace, nan, trapz, zeros_like
from scipy.interpolate import splev, splrep
from scipy.linalg import solveh_banded

from radis.misc.arrays import (
    anynan,
    first_nonnan_index,
    is_sorted,
    is_sorted_backward,
    last_nonnan_index,
)
from radis.misc.debug import printdbg


def resample(
    xspace,
    vector,
    xspace_new,
    k=1,
    ext="error",
    energy_threshold="default",
    print_conservation=True,
):
    """Resample (xspace, vector) on a new space (xspace_new) of evenly
    distributed data and whose bounds are taken as the same as `xspace`.

    Uses spline interpolation to create the intermediary points. Number of points
    is the same as the initial xspace, times a resolution factor. Verifies energy
    conservation on the intersecting range at the end.


    Parameters
    ----------
    xspace: array
        space on which vector was generated
    vector: array
        quantity to resample
    xspace_new: array
        space on which to resample

    Other Parameters
    ----------------
    resfactor: array
        xspace vector to resample on
    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.
    ext: 'error', 'extrapolate', 0, 1
        Controls the value returned for elements of xspace_new not in the interval
        defined by xspace. If 'error', raise a ValueError. If 'extrapolate', well,
        extrapolate. If '0' or 0, then fill with 0. If 1, fills with 1.
        Default 'error'.
    energy_threshold: float or ``None`` or ``'default'
        if energy conservation (integrals on the intersecting range) is above
        this threshold, raise an error. If ``None``, dont check for energy conservation.
        If ``'default'``, look up the value in :py:attr:`radis.config` ["RESAMPLING_TOLERANCE_THRESHOLD"]
        Default ``'default'``
    print_conservation: boolean
        if True, prints energy conservation


    Returns
    -------
    array: resampled vector on evenly spaced array. Number of element is conserved.

    Notes
    -----
    Note that depending upon the from_space > to_space operation, sorting may
    be reversed.

    Examples
    --------
    Resample a :class:`~radis.spectrum.spectrum.Spectrum` radiance
    on an evenly spaced wavenumber space::

        w_nm, I_nm = s.get('radiance')
        w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm)

    See Also
    --------
    :py:meth:`~radis.spectrum.spectrum.Spectrum.resample`
    """

    if len(xspace) != len(vector):
        raise ValueError(
            "vector and xspace should have the same length. "
            + "Got {0}, {1}".format(len(vector), len(xspace))
        )
    if energy_threshold == "default":
        import radis

        energy_threshold = radis.config["RESAMPLING_TOLERANCE_THRESHOLD"]

    # Check reversed (interpolation requires objects are sorted)
    if is_sorted(xspace):
        reverse = False
    elif is_sorted_backward(xspace):
        reverse = True
    else:
        raise ValueError(
            "Resampling requires wavespace to be sorted. It is not! If working with a Spectrum object, use `s.sort()`? "
        )

    if reverse:
        xspace = xspace[::-1]
        xspace_new = xspace_new[::-1]
        vector = vector[::-1]

    # translate ext in FITPACK syntax for splev
    if ext == "extrapolate":
        ext_fitpack = 0  # splev returns extrapolated value
    elif ext in [0, "0", 1, "1", nan, "nan"]:
        ext_fitpack = 1  # splev returns 0  (fixed in post-processing)
    elif ext == "error":
        ext_fitpack = 2  # splev raises ValueError
    else:
        raise ValueError("Unexpected value for `ext`: {0}".format(ext))

    # Handle nans in vector :
    # FITPACK will fail if there are nans in the vector being interpolated.
    # here we trim the sides; and raise an error if there are nans in the middle
    m = first_nonnan_index(vector)
    M = last_nonnan_index(vector)
    if m is None:
        raise ValueError("Vector being resample has only nans. Check your data")
    nanmask = False
    if m > 0 or M < len(vector) - 1:  # else, no change has to be made
        nanmask = zeros_like(vector, dtype=bool)
        nanmask[m : M + 1] = True
        xspace = xspace[nanmask]
        vector = vector[nanmask]
    if anynan(vector):
        raise ValueError(
            "Vector being resample has {0} nans. Interpolation will fail".format(
                isnan(vector).sum()
            )
        )

    # Resample the slit function on the spectrum grid
    try:
        tck = splrep(xspace, vector, k=k)
    except ValueError:
        # Probably error on input data. Print it before crashing.
        print("\nValueError - Input data below:")
        print("-" * 5)
        print(xspace)
        print(vector)
        print("Check plot 101 too")
        import matplotlib.pyplot as plt

        plt.figure(101).clear()
        plt.plot(xspace, vector)
        plt.xlabel("xspace")
        plt.ylabel("vector")
        plt.title("ValueError")
        raise
    vector_new = splev(xspace_new, tck, ext=ext_fitpack)

    # ... get masks
    b = (xspace >= xspace_new.min()) * (xspace <= xspace_new.max())
    b_new = (xspace_new >= xspace.min()) * (xspace_new <= xspace.max())

    # fix filling for out of boundary values
    if ext in [1, "1"]:
        vector_new[~b_new] = 1
        if __debug__:
            printdbg(
                "Filling with 1 on w<{0}, w>{1} ({2} points)".format(
                    xspace.min(), xspace.max(), (1 - b_new).sum()
                )
            )
    elif ext in [nan, "nan"]:
        vector_new[~b_new] = nan
        if __debug__:
            printdbg(
                "Filling with nans on w<{0}, w>{1} ({2} points)".format(
                    xspace.min(), xspace.max(), (1 - b_new).sum()
                )
            )

    # Check energy conservation:

    # ... calculate energy
    energy0 = abs(trapz(vector[b], x=xspace[b]))
    energy_new = abs(trapz(vector_new[b_new], x=xspace_new[b_new]))
    if energy_new == 0:  # deal with particular case of energy = 0
        if energy0 == 0:
            energy_ratio = 1
        else:
            energy_ratio = 0
    else:  # general case
        energy_ratio = energy0 / energy_new
    if energy_threshold:
        if abs(energy_ratio - 1) > energy_threshold:
            import matplotlib.pyplot as plt

            plt.figure(101).clear()
            plt.plot(xspace, vector, "-ok", label="original")
            plt.plot(xspace_new, vector_new, "-or", label="resampled")
            plt.xlabel("xspace")
            plt.ylabel("vector")
            plt.legend()
            raise ValueError(
                "Error in resampling: "
                + "difference in areas (i.e. energy conservation ) ({0:.5g}%) is above the tolerance level ({1:.5g}%)".format(
                    (1 - energy_ratio) * 100, energy_threshold * 100
                )
                + ". Check graph 101. If resampling a low resolution spectrum on a high resolution spectrum, try the other way around. "
                + "Increasing the tolerance threshold is possible but may result in accuracy errors. To increase the tolerance, set `resample(..., energy_threshold=...)`, or change the default global value in `radis.config['RESAMPLING_TOLERANCE_THRESHOLD']`"
            )
    if print_conservation:
        print("Resampling - Energy conservation: {0:.5g}%".format(energy_ratio * 100))

    # Reverse again
    if reverse:
        # xspace_new = xspace_new[::-1]
        vector_new = vector_new[::-1]

    return vector_new


def resample_even(
    xspace,
    vector,
    resfactor=2,
    k=1,
    ext="error",
    energy_threshold=1e-3,
    print_conservation=True,
):
    """Resample (xspace, vector) on a new space (xspace_new) of evenly
    distributed data and whose bounds are taken as the same as `xspace`.

    Uses spline interpolation to create the intermediary points. Number of points
    is the same as the initial xspace, times a resolution factor. Verifies energy
    conservation at the end.


    Parameters
    ----------
    xspace: array
        space on which vector was generated
    vector: array
        quantity to resample
    resfactor: float
        increase of resolution. If 1, output vector has the same number of
        points as the input vector. Default 2.
    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.
    ext: 'error', 'extrapolate', 0
        Controls the value returned for elements of xspace_new not in the interval
        defined by xspace. If 'error', raise a ValueError. If 'extrapolate', well,
        extrapolate. If '0' or 0, then fill with 0. Default 'error'.
    energy_threshold: float
        if energy conservation (integrals) is above this threshold, raise an
        error
    print_conservation: boolean
        if True, prints energy conservation


    Returns
    -------
    xspace_new: array
        evenly spaced mapping of xspace (same min, same max)
    vector_new: array
        resampled vector on evenly spaced array. Number of element is conserved.

    Note that depending upon the from_space > to_space operation, sorting may
    be reversed.


    Examples
    --------
    Resample a :class:`~radis.spectrum.spectrum.Spectrum` radiance
    on an evenly spaced wavenumber space::

        w_nm, I_nm = s.get('radiance')
        w_cm, I_cm = resample_even(nm2cm(w_nm), I_nm)

    You can also use :py:meth:`~radis.spectrum.spectrum.Spectrum.resample_even`
    directly
    """

    # Get new evenly space array
    xspace_new = linspace(xspace[0], xspace[-1], len(vector) * resfactor)

    # Resample
    vector_new = resample(
        xspace,
        vector,
        xspace_new,
        k=k,
        ext=ext,
        energy_threshold=energy_threshold,
        print_conservation=print_conservation,
    )

    return xspace_new, vector_new


def als_baseline(
    intensities,
    asymmetry_param=0.05,
    smoothness_param=1e6,
    max_iters=10,
    conv_thresh=1e-5,
    verbose=False,
):
    """Computes the asymmetric least squares baseline

    Parameters
    ----------
    intensities: array_like
        vector to smooth
    asymmetry_param: float
        value will shift the baseline fit below or above your average line. To cancel gaussian noise,
        you would want the value to be ``0.5``. To remove a positive peak would want it to be close
        to ``0``. To remove a negative peaks (from a transmittance signal for instance) you would
        want it to be close to ``1``. Default ``0.05``
    smoothness_param: float
        Relative importance of smoothness of the predicted response:
        the higher, the smoother the baseline. Suggested range: ``1e2`` to ``1e8``. Default ``1e6``

    Other Parameters
    ----------------
    max_iters: int
        number of iterations
    conv_thresh: float
        convergence
    verbose: boolean


    Examples
    --------
    ::

        from neq.math.smooth import als_baseline
        I = als_baseline(I, smoothness_param=1e5, asymmetry_param=0.1)

    References
    ----------

    * https://zanran_storage.s3.amazonaws.com/www.science.uva.nl/ContentPages/443199618.pdf
    * http://www.science.uva.nl/~hboelens/publications/draftpub/Eilers_2005.pdf

    Notes
    -----

    Implementation:

    - uses :func:`~neq.math.smooth.WhittakerSmoother` internally
    - kudos to https://gist.github.com/perimosocordiae

    See Also
    --------

    :func:`~radis.misc.signal.WhittakerSmoother`

    """
    smoother = WhittakerSmoother(intensities, smoothness_param, deriv_order=2)
    # Rename p for concision.
    p = asymmetry_param
    # Initialize weights.
    w = np.ones(intensities.shape[0])
    for i in range(max_iters):
        z = smoother.smooth(w)
        mask = intensities > z
        new_w = p * mask + (1 - p) * (~mask)
        conv = np.linalg.norm(new_w - w)
        if verbose:
            print(i + 1, conv)
        if conv < conv_thresh:
            break
        w = new_w
    else:
        print("ALS did not converge in %d iterations" % max_iters)
    return z


def baseline(y, deg=None, max_it=None, tol=None):
    """
    Computes the baseline of a given data.

    Iteratively performs a polynomial fitting in the data to detect its
    baseline. At every iteration, the fitting weights on the regions with
    peaks are reduced to identify the baseline only.

    Parameters
    ----------
    y : ndarray
        Data to detect the baseline.
    deg : int (default: 3)
        Degree of the polynomial that will estimate the data baseline. A low
        degree may fail to detect all the baseline present, while a high
        degree may make the data too oscillatory, especially at the edges.
    max_it : int (default: 100)
        Maximum number of iterations to perform.
    tol : float (default: 1e-3)
        Tolerance to use when comparing the difference between the current
        fit coefficients and the ones from the last iteration. The iteration
        procedure will stop when the difference between them is lower than
        *tol*.

    Returns
    -------
    ndarray
        Array with the baseline amplitude for every original point in *y*

    References
    -------
    This function has been taken from the PeakUtils package.
    Their source code can be found here: https://bitbucket.org/lucashnegri/peakutils
    More info: https://peakutils.readthedocs.io/en/latest/
    """
    # for not repeating ourselves in `envelope`
    if deg is None:
        deg = 3
    if max_it is None:
        max_it = 100
    if tol is None:
        tol = 1e-3

    order = deg + 1
    coeffs = np.ones(order)

    # try to avoid numerical issues
    cond = math.pow(abs(y).max(), 1.0 / order)
    x = np.linspace(0.0, cond, y.size)
    base = y.copy()

    vander = np.vander(x, order)
    vander_pinv = LA.pinv(vander)

    for _ in range(max_it):
        coeffs_new = np.dot(vander_pinv, y)

        if LA.norm(coeffs_new - coeffs) / LA.norm(coeffs) < tol:
            break

        coeffs = coeffs_new
        base = np.dot(vander, coeffs)
        y = np.minimum(y, base)

    return base


class WhittakerSmoother(object):
    """

    References
    ----------
    - kudos to https://gist.github.com/perimosocordiae

    See Also
    --------

    :func:`~radis.misc.signal.als_baseline`
    """

    def __init__(self, signal, smoothness_param, deriv_order=1):
        self.y = signal
        assert deriv_order > 0, "deriv_order must be an int > 0"
        # Compute the fixed derivative of identity (D).
        d = np.zeros(deriv_order * 2 + 1, dtype=int)
        d[deriv_order] = 1
        d = np.diff(d, n=deriv_order)
        n = self.y.shape[0]
        k = len(d)
        s = float(smoothness_param)

        # Here be dragons: essentially we're faking a big banded matrix D,
        # doing s * D.T.dot(D) with it, then taking the upper triangular bands.
        diag_sums = np.vstack(
            [
                np.pad(s * np.cumsum(d[-i:] * d[:i]), ((k - i, 0),), "constant")
                for i in range(1, k + 1)
            ]
        )
        upper_bands = np.tile(diag_sums[:, -1:], n)
        upper_bands[:, :k] = diag_sums
        for i, ds in enumerate(diag_sums):
            upper_bands[i, -i - 1 :] = ds[::-1][: i + 1]
        self.upper_bands = upper_bands

    def smooth(self, w):
        foo = self.upper_bands.copy()
        foo[-1] += w  # last row is the diagonal

        return solveh_banded(foo, w * self.y, overwrite_ab=True, overwrite_b=True)


if __name__ == "__main__":
    from radis.test.spectrum.test_spectrum import (
        test_resampling_function,
        test_resampling_nan_function,
    )

    test_resampling_function()
    test_resampling_nan_function()



# -*- coding: utf-8 -*-
"""

"""

from radis.db.molparam import MOLPARAMS_EXTRA_PATH, MolParams


def test_molparams(verbose=True, *args, **kwargs):

    molpar = MolParams()

    assert molpar.get(2, 2, "abundance") == 0.0110574
    assert molpar.get("CO2", 1, "molar_mass") == 43.98983

    # Test errors :
    import pytest

    with pytest.raises(NotImplementedError):
        molpar.get("C20H25N3O", 1, "molar_mass")

    # Test reading of extra parameters
    import pytest

    molpar = MolParams(extra_file_json=None)
    with pytest.raises(NotImplementedError):
        molpar.get("trans-P2H2", 1, "molar_mass") == 43.98983

    molpar = MolParams(extra_file_json=MOLPARAMS_EXTRA_PATH)
    assert "abundance" in molpar.extra_molparams
    assert "molar_mass" in molpar.extra_molparams
    assert molpar.get("trans-P2H2", 1, "molar_mass") == 63.9631740604


if __name__ == "__main__":
    test_molparams()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Handy functions for manipulation of :class:`~radis.spectrum.spectrum.Spectrum` objects

Routine Listing
---------------

Operators:

- :py:func:`~radis.spectrum.operations.multiply` : simply use ``s*2``
- :py:func:`~radis.spectrum.operations.add_constant` : simply use ``s+1``
- :py:func:`~radis.spectrum.operations.add_array` : simply use ``s+a``
- :py:func:`~radis.spectrum.operations.add_spectra` : simply use ``s1+s2``
- :py:func:`~radis.spectrum.operations.substract_spectra` : simply use ``s1-s2``

Note that these operators are purely algebraic and should not be used in place
of the line-of-sight functions, i.e, :py:func:`~radis.los.slabs.SerialSlabs` (``>``)
and :py:func:`~radis.los.slabs.MergeSlabs` (``//``)

Functions to manipulate one spectrum:

- :py:func:`~radis.spectrum.operations.crop`
- :py:func:`~radis.spectrum.operations.offset`

Play with baselines:

- :py:func:`~radis.spectrum.operations.get_baseline`
- :py:func:`~radis.spectrum.operations.sub_baseline`

Functions to discard all but one spectral quantity:

- :py:func:`~radis.spectrum.operations.Transmittance`
- :py:func:`~radis.spectrum.operations.Transmittance_noslit`
- :py:func:`~radis.spectrum.operations.Radiance`
- :py:func:`~radis.spectrum.operations.Radiance_noslit`

Keeps all spectral quantities, but make emission equal to 0 (useful when
calculating contributions of line of sight slabs):

- :py:func:`~radis.spectrum.operations.PerfectAbsorber`


Examples
--------

Most of these functions are implemented with the standard operators. Ex::

    ((s_exp - 0.1)*10).plot()   # works for a Spectrum s_exp

-------------------------------------------------------------------------------


"""


from warnings import warn

import astropy.units as u
from numpy import hstack, ones_like

from radis.phys.convert import (
    air2vacuum,
    cm2nm,
    cm2nm_air,
    dcm2dnm,
    dcm2dnm_air,
    dnm2dcm,
    dnm_air2dcm,
    nm2cm,
    nm_air2cm,
    vacuum2air,
)
from radis.phys.units import Unit
from radis.phys.units_astropy import convert_and_strip_units
from radis.spectrum.spectrum import Spectrum

# %% Filter Spectra


def Transmittance(s: Spectrum) -> Spectrum:
    """Returns a new Spectrum with only the ``transmittance`` component of ``s``

    Parameters
    ----------
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object

    Returns
    -------
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only the ``transmittance``,
        ``absorbance`` and/or ``abscoeff`` part of ``s``, where ``radiance_noslit`` ,
        ``emisscoeff`` and ``emissivity_noslit`` (if they exist) have been set to 0

    Examples
    --------
    This function is useful to use :ref:`Spectrum algebra <label_spectrum_algebra>`
    operations::

        s = calc_spectrum(...)   # contains emission & absorption arrays
        tr = Transmittance(s)    # contains 'radiance_noslit' array only
        tr -= 0.1    # arithmetic operation is applied to Transmittance only

    Equivalent to::

        rad = s.take('transmittance')

    See Also
    --------
    :py:func:`~radis.spectrum.operations.Transmittance_noslit`,
    :py:func:`~radis.spectrum.operations.Radiance_noslit`,
    :py:func:`~radis.spectrum.operations.Radiance`,
    :py:meth:`~radis.spectrum.Spectrum.take`

    """
    warn(
        "Transmittance(Spectrum) will be deprecated in future RADIS versions. Use Spectrum.take('transmittance')"
    )

    return s.copy(copy_lines=True, quantity="transmittance")


def Transmittance_noslit(s: Spectrum) -> Spectrum:
    """Returns a new Spectrum with only the ``transmittance_noslit`` component of ``s``

    Parameters
    ----------
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object

    Returns
    -------
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only
        ``transmittance_noslit`` defined

    Examples
    --------
    This function is useful to use :ref:`Spectrum algebra <label_spectrum_algebra>`
    operations::

        s = calc_spectrum(...)   # contains emission & absorption arrays
        tr = Transmittance_noslit(s) # contains 'radiance_noslit' array only
        tr -= 0.1    # arithmetic operation is applied to Transmittance_noslit only

    Equivalent to::

        rad = s.take('transmittance_noslit')

    See Also
    --------
    :py:func:`~radis.spectrum.operations.Transmittance`
    :py:func:`~radis.spectrum.operations.Radiance_noslit`,
    :py:func:`~radis.spectrum.operations.Radiance`,
    :py:meth:`~radis.spectrum.Spectrum.take`

    """
    warn(
        "Transmittance_noslit(Spectrum) will be deprecated in future RADIS versions. Use Spectrum.take('transmittance_noslit')"
    )

    return s.copy(copy_lines=True, quantity="transmittance_noslit")


def Radiance(s: Spectrum) -> Spectrum:
    """Returns a new Spectrum with only the ``radiance`` component of ``s``

    Parameters
    ----------
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object

    Returns
    -------
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only ``radiance``
        defined

    Examples
    --------
    This function is useful to use :ref:`Spectrum algebra <label_spectrum_algebra>`
    operations::

        s = calc_spectrum(...)   # contains emission & absorption arrays
        rad = Radiance(s)        # contains radiance array only
        rad -= 0.1    # arithmetic operation is applied to Radiance only

    Equivalent to::

        rad = s.take('radiance')

    See Also
    --------
    :py:func:`~radis.spectrum.operations.Radiance_noslit`
    :py:func:`~radis.spectrum.operations.Transmittance_noslit`,
    :py:func:`~radis.spectrum.operations.Transmittance`,
    :py:meth:`~radis.spectrum.Spectrum.take`

    """
    warn(
        "Radiance(Spectrum) will be deprecated in future RADIS versions. Use Spectrum.take('radiance')"
    )

    return s.copy(copy_lines=True, quantity="radiance")


def Radiance_noslit(s: Spectrum) -> Spectrum:
    """Returns a new Spectrum with only the ``radiance_noslit`` component of ``s``

    Parameters
    ----------
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object

    Returns
    -------
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only ``radiance_noslit``
        defined

    Examples
    --------
    This function is useful to use :ref:`Spectrum algebra <label_spectrum_algebra>`
    operations::

        s = calc_spectrum(...)   # contains emission & absorption arrays
        rad = Radiance_noslit(s) # contains 'radiance_noslit' array only
        rad -= 0.1    # arithmetic operation is applied to Radiance_noslit only

    Equivalent to::

        rad = s.take('radiance_noslit')

    See Also
    --------
    :py:func:`~radis.spectrum.operations.Radiance`,
    :py:func:`~radis.spectrum.operations.Transmittance_noslit`,
    :py:func:`~radis.spectrum.operations.Transmittance`,
    :py:meth:`~radis.spectrum.Spectrum.take`

    """
    warn(
        "Radiance_noslit(Spectrum) will be deprecated in future RADIS versions. Use Spectrum.take('radiance_noslit')"
    )

    return s.copy(copy_lines=True, quantity="radiance_noslit")


# Useful for determining line-of-sight contributions:


def PerfectAbsorber(s: Spectrum) -> Spectrum:
    """Makes a new Spectrum with the same transmittance/absorbance as Spectrum
    ``s``, but with radiance set to 0.
    Useful to get contribution of different slabs in line-of-sight
    calculations (see example).

    Parameters
    ----------
    s: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object

    Returns
    -------
    s_tr: Spectrum
        :class:`~radis.spectrum.spectrum.Spectrum` object, with only the ``transmittance``,
        ``absorbance`` and/or ``abscoeff`` part of ``s``, where ``radiance_noslit`` ,
        ``emisscoeff`` and ``emissivity_noslit`` (if they exist) have been set to 0

    Examples
    --------
    Let's say you have a total line of sight::

        s_los = s1 > s2 > s3

    If you now want to get the contribution of ``s2`` to the line-of-sight emission,
    you can do::

        (s2 > PerfectAbsorber(s3)).plot('radiance_noslit')

    And the contribution of ``s1`` would be::

        (s1 > PerfectAbsorber(s2>s3)).plot('radiance_noslit')

    See more examples in :ref:`Line-of-Sight module <label_los_index>`
    """

    s_tr = s.copy()
    for k in ["radiance_noslit", "emisscoeff", "emissivity_noslit"]:
        if k in s_tr._q:
            s_tr._q[k] *= 0

    for k in ["radiance", "emissivity"]:
        if k in s_tr._q:
            s_tr._q[k] *= 0

    # Deactivate equilibrium conditions (so Kirchoff's law cannot be used anymore)
    s_tr.conditions["thermal_equilibrium"] = False

    s_tr.name = "PureAbsorber({0})".format(s.get_name())

    return s_tr


# %% Change wavelength


def crop(s: Spectrum, wmin=None, wmax=None, wunit=None, inplace=False) -> Spectrum:
    # type: (Spectrum, float, float, str, str, bool) -> Spectrum
    """Crop spectrum to ``wmin-wmax`` range in ``wunit``

    Parameters
    ----------
    s: Spectrum object
        object to crop
    wmin, wmax: float, or None
        boundaries of spectral range (in ``wunit``)

        wunit: ``'nm'``, ``'cm-1'``, ``'nm_vac'``
            which waveunit to use for ``wmin, wmax``. If ``default``:
            use the default Spectrum wavespace defined with
            :meth:`~radis.spectrum.spectrum.Spectrum.get_waveunit`.

    Other Parameters
    ----------------
    inplace: bool
        if ``True``, modify ``s`` directly. Else, returns a copy.

    Returns
    -------
    s_crop: Spectrum
        a cropped Spectrum.
        if using ``inplace``, then ``s_crop`` and ``s`` are still the same object

    Examples
    --------
    ::

        crop(s, 420, 480, 'nm', 'air')
    Or in ``cm-1``::

        crop(s, 2000, 2300, 'cm-1')
    """

    stored_waveunit = s.get_waveunit()

    wlunit_list = {
        "km": u.km,
        "m": u.m,
        "cm": u.cm,
        "mm": u.mm,
        "um": u.um,
    }

    # Check inputs

    if wmin is None and wmax is None:
        raise ValueError("Choose at least `wmin=` or `wmax=`")

    if (
        type(wmin) == type(wmax) == u.quantity.Quantity
    ):  # User states astropy units directly to wmin and wmax
        # Convert wmin and wmax to nm unit for later Spectrum wavespace conversion
        # The stored waveunit can only be nm or cm-1, but we don't need to deal with wavenumber unit conversion, only wavelength
        # So I convert them all to nm
        wmin = convert_and_strip_units(wmin, u.nm)
        wmax = convert_and_strip_units(wmax, u.nm)

    if wunit is None:
        raise ValueError("Please precise unit for wmin and wmax with `unit=`")
    elif wunit not in ["nm", "cm-1"]:
        if wunit in wlunit_list:
            wmin = convert_and_strip_units(wmin * wlunit_list[wunit], u.nm)
            wmax = convert_and_strip_units(wmax * wlunit_list[wunit], u.nm)
            wunit = "nm"
        else:
            raise ValueError(
                f"Unsupported wunit, should be cm-1, nm or {wlunit_list.keys}"
            )
    # assert wunit in ["nm", "cm-1"]

    if (wmin is not None and wmax is not None) and wmin >= wmax:
        raise ValueError(
            "wmin should be < wmax (Got: {0:.2f}, {1:.2f})".format(wmin, wmax)
        )

    if not inplace:
        s = s.copy()

    # Convert wmin, wmax to Spectrum wavespace  (stored_waveunit)
    # (deal with cases where wavelength are given in 'air' or 'vacuum')
    # TODO @dev: rewrite with wunit='cm-1', 'nm_air', 'nm_vac'
    wmin0, wmax0 = wmin, wmax
    if stored_waveunit == "cm-1":
        # convert wmin, wmax to wavenumber
        if wunit == "nm":
            if wmax0:
                wmin = nm_air2cm(wmax0)  # note: min/max inverted
            if wmin0:
                wmax = nm_air2cm(wmin0)  # note: min/max inverted
        elif wunit == "nm_vac":
            if wmax0:
                wmin = nm2cm(wmax0)  # note: min/max inverted
            if wmin0:
                wmax = nm2cm(wmin0)  # note: min/max inverted
        elif wunit == "cm-1":
            pass
        else:
            raise ValueError(wunit)
    elif stored_waveunit == "nm":
        # convert wmin, wmax to wavelength air
        if wunit == "nm":
            pass
        elif wunit == "nm_vac":
            if wmin0:
                wmin = vacuum2air(wmin0)
            if wmax0:
                wmax = vacuum2air(wmax0)
        elif wunit == "cm-1":
            if wmax0:
                wmin = cm2nm_air(wmax0)  # note: min/max inverted
            if wmin0:
                wmax = cm2nm_air(wmin0)  # note: min/max inverted
        else:
            raise ValueError(wunit)
    elif stored_waveunit == "nm_vac":
        # convert wmin, wmax to wavelength vacuum
        if wunit == "nm":
            if wmin0:
                wmin = air2vacuum(wmin0)
            if wmax0:
                wmax = air2vacuum(wmax0)
        elif wunit == "nm_vac":
            pass
        elif wunit == "cm-1":
            if wmax0:
                wmin = cm2nm(wmax0)  # note: min/max inverted
            if wmin0:
                wmax = cm2nm(wmin0)  # note: min/max inverted
        else:
            raise ValueError(wunit)
    else:
        raise ValueError(stored_waveunit)

    # Crop
    if len(s._q) > 0:
        b = ones_like(s._q["wavespace"], dtype=bool)
        if wmin:
            b *= wmin <= s._q["wavespace"]
        if wmax:
            b *= s._q["wavespace"] <= wmax
        for k, v in s._q.items():
            s._q[k] = v[b]

    return s


# %% Algebraic operations on Spectra


def _get_unique_var(s, var, inplace):
    """Returns the unique spectral quantity in the Spectrum ``s``. If there are more,
    raises an error.
    """
    # If var is undefined, get it if there is no ambiguity
    if var is None:
        var = s._get_unique_var()
    # If inplace, assert there is only one var
    if inplace and len(s.get_vars()) > 1:
        raise ValueError(
            "Cant modify inplace one spectral quantity of a Spectrum "
            + "that contains several ({0}). Use 'inplace=False'".format(s.get_vars())
        )
    return var


def multiply(s, coef, unit=None, var=None, inplace=False):
    """Multiply s[var] by the float 'coef'

    Parameters
    ----------
    s: Spectrum object
        The spectrum to multiply.
    coef: float
        Coefficient of the multiplication.
    unit: str, or `~astropy.units.core.Unit`
        unit for ``coef``. If ``None``, ``coef`` is considered to be
        adimensioned. Else, the spectrum `~radis.spectrum.spectrum.Spectrum.units` is multiplied.
    var: str, or ``None``
        'radiance', 'transmittance', ... If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy.
        Default ``False``

    Returns
    -------
    s : Spectrum
        Spectrum object where intensity of s['var'] is multiplied by coef
        If ``inplace=True``, ``s`` has been modified directly.

    """
    # Check input
    var = _get_unique_var(s, var, inplace)

    # Case where a is dimensioned
    if isinstance(coef, u.quantity.Quantity):
        if unit is not None:
            raise ValueError(
                "Cannot use unit= when giving a dimensioned array ({0})".format(
                    coef.unit
                )
            )
        unit = coef.unit
        coef = coef.value

    if not inplace:
        s = s.copy(quantity=var)

    # Multiply inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), Iunit=s.units[var], copy=False)
    I *= coef  # @dev: updates the Spectrum directly because of copy=False

    # Convert Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        s.units[var] = (Unit(Iunit) * Unit(unit)).to_string()

    return s


def add_constant(s, cst, unit=None, var=None, inplace=False):
    """Return a new spectrum with a constant added to s[var].
    Equivalent to::

        s + constant

    Parameters
    ----------
    s: Spectrum objects
        Spectrum you want to modify
    cst: float
        Constant to add.
    unit: str
        unit for ``cst``. If ``None``, uses the default unit in ``s`` for
        variable ``var``.
    var: str, or ``None``
        'radiance', 'transmittance', ... If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy.
        Default ``False``

    Returns
    -------
    s : Spectrum
        Spectrum object where cst is added to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    Notes
    -----

    Use only for rough work. If you want to work properly with spectrum
    objects, see :py:meth:`~radis.los.slabs.MergeSlabs`.
    """
    # Check input
    var = _get_unique_var(s, var, inplace)

    # Note: using dimensioned value for `cst` will make it an array,
    # processed by add_array

    # Convert to Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        if unit != Iunit:
            from radis.phys.convert import conv2

            cst = conv2(cst, unit, Iunit)

    if not inplace:
        s = s.copy(quantity=var)

    # Add inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), Iunit=s.units[var], copy=False)
    I += cst
    # @dev: updates the Spectrum directly because of copy=False

    #    s.name = '{0}+{1}'.format(s.get_name(), cst)

    return s


def add_array(s, a, unit=None, var=None, inplace=False):
    """Return a new spectrum with a constant added to s[var].
    Equivalent to::

        s + array

    Parameters
    ----------
    s: Spectrum objects
        Spectrum you want to modify
    a: numpy array, or `~astropy.units.quantity.Quantity`
        array to add. Must have the same length as variable ``var`` in Spectrum
        ``s``. Can be dimensioned with :py:mod:`~astropy.units`.
    unit: str
        unit for ``a``. If ``None``, uses the default unit in ``s`` for
        variable ``var``.
    var: str, or ``None``
        'radiance', 'transmittance', ... If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy.
        Default ``False``

    Returns
    -------
    s : Spectrum
        Spectrum object where array ``a`` is added to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    Notes
    -----
    Use only for rough work. If you want to work properly with spectrum
    objects, see MergeSlabs.

    Examples
    --------

    Add Gaussian noise to your Spectrum (assuming there is only one spectral
    quantity defined)::

        s += np.random.normal(0,1,len(s.get_wavelength()))

    """
    # Check input
    var = _get_unique_var(s, var, inplace)

    # Case where a is dimensioned
    if isinstance(a, u.quantity.Quantity):
        if unit is not None:
            raise ValueError(
                "Cannot use unit= when giving a dimensioned array ({0})".format(a.unit)
            )
        unit = a.unit
        a = a.value

    # Convert to Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        if unit != Iunit:
            from radis.phys.convert import conv2

            a = conv2(a, unit, Iunit)

    if not inplace:
        s = s.copy(quantity=var)

    # Add inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), Iunit=s.units[var], copy=False)
    I += a
    # @dev: updates the Spectrum directly because of copy=False

    return s


def sub_baseline(s, left, right, unit=None, var=None, inplace=False):
    """Return a new spectrum with a baseline subtracted to s[var]

    Parameters
    ----------

    s: Spectrum objects
        Spectrum you want to modify
    left: Float
        Constant to substract on the left of the spectrum.
    right: Float
        Constant to substract on the right of the spectrum.
    unit: str
        unit for ``cst``. If ``None``, uses the default unit in ``s`` for
        variable ``var``.
    var: str
        'radiance', 'transmittance', ...  If ``None``, get the unique spectral
        quantity of ``s`` or raises an error if there is any ambiguity
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy.
        Default ``False``

    Returns
    -------

    s: Spectrum
        Spectrum object where the baseline was subtracted to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    Notes
    -----
    Use only for rough work.

    See Also
    --------

    :py:func:`~radis.spectrum.operations.get_baseline`
    """
    import numpy as np

    # Check input
    var = _get_unique_var(s, var, inplace)

    # Case where left, right are dimensioned
    if isinstance(left, u.quantity.Quantity) or isinstance(right, u.quantity.Quantity):
        if unit is not None:
            raise ValueError(
                "Cannot use unit= when giving a dimensioned array ({0})".format(
                    left.unit
                )
            )
        if not isinstance(left, u.quantity.Quantity) or not isinstance(
            right, u.quantity.Quantity
        ):
            raise ValueError(
                "Both left and right arguments must be dimensioned with the same unit"
            )
        if left.unit != right.unit:
            raise ValueError(
                "Both left and right arguments must be dimensioned with the same unit"
            )
        unit = left.unit
        left = left.value
        right = right.value

    # Convert to Spectrum unit
    if unit is not None:
        Iunit = s.units[var]
        if unit != Iunit:
            from radis.phys.convert import conv2

            left = conv2(left, unit, Iunit)
            right = conv2(right, unit, Iunit)

    if not inplace:
        s = s.copy(quantity=var)

    # Subtract inplace       ( @dev: we have copied already if needed )
    w, I = s.get(var, wunit=s.get_waveunit(), Iunit=s.units[var], copy=False)
    I -= np.linspace(left, right, num=np.size(I))
    # @dev: updates the Spectrum directly because of copy=False

    return s


def add_spectra(s1, s2, var=None, force=False):
    """Return a new spectrum with ``s2`` added to ``s1``.
    Equivalent to::

        s1 + s2

    .. warning::
        we are just algebraically adding the quantities. If you want to merge
        spectra while preserving the radiative transfer equation, see
        :func:`~radis.los.slabs.MergeSlabs` and :func:`~radis.los.slabs.SerialSlabs`

    Parameters
    ----------

    s1, s2: Spectrum objects
        Spectrum you want to substract
    var: str
        quantity to manipulate: 'radiance', 'transmittance', ... If ``None``,
        get the unique spectral quantity of ``s1``, or the unique spectral
        quantity of ``s2``, or raises an error if there is any ambiguity

    Returns
    -------

    s: Spectrum
        Spectrum object with the same units and waveunits as ``s1``

    See Also
    --------

    :func:`~radis.los.slabs.MergeSlabs`,
    :func:`~radis.spectrum.operations.substract_spectra`

    """

    # Get variable
    if var is None:
        try:
            var = _get_unique_var(
                s2, var, inplace=False
            )  # unique variable of 2nd spectrum
        except KeyError:
            var = _get_unique_var(
                s1, var, inplace=False
            )  # if doesnt exist, unique variable of 1st spectrum
            # if it fails, let it fail
    # Make sure it is in both Spectra
    if var not in s1.get_vars():
        raise KeyError("Variable {0} not in Spectrum {1}".format(var, s1.get_name()))
    if var not in s2.get_vars():
        raise KeyError("Variable {0} not in Spectrum {1}".format(var, s1.get_name()))

    if var in ["transmittance_noslit", "transmittance"] and not force:
        raise ValueError(
            "It does not make much physical sense to sum transmittances. Are "
            + "you sure of what you are doing? See also `//` (MergeSlabs), `>` "
            + "(SerialSlabs) and `concat_spectra`. If you're sure, use `force=True`"
        )

    # Get s1 units
    Iunit1 = s1.units[var]
    wunit1 = s1.get_waveunit()

    # Resample s2 on s1
    s2 = s2.resample(s1, inplace=False)

    # Add, change output unit if needed.
    w1, I1 = s1.get(var=var, Iunit=Iunit1, wunit=wunit1)
    w2, I2 = s2.get(var=var, Iunit=Iunit1, wunit=wunit1)

    name = s1.get_name() + "+" + s2.get_name()

    sub = Spectrum.from_array(w1, I1 + I2, var, wunit=wunit1, Iunit=Iunit1, name=name)
    #    warn("Conditions of the left spectrum were copied in the substraction.", Warning)
    return sub


def substract_spectra(s1, s2, var=None):
    """Return a new spectrum with ``s2`` subtracted from ``s1``.
    Equivalent to::

        s1 - s2

    Parameters
    ----------

    s1, s2: Spectrum objects
        Spectrum you want to substract
    var: str
        quantity to manipulate: 'radiance', 'transmittance', ... If ``None``,
        get the unique spectral quantity of ``s1``, or the unique spectral
        quantity of ``s2``, or raises an error if there is any ambiguity

    Returns
    -------

    s: Spectrum
        Spectrum object with the same units and waveunits as ``s1``

    See Also
    --------

    :func:`~radis.spectrum.operations.add_spectra`

    """

    # Get variable
    if var is None:
        try:
            var = _get_unique_var(
                s2, var, inplace=False
            )  # unique variable of 2nd spectrum
        except KeyError:
            var = _get_unique_var(
                s1, var, inplace=False
            )  # if doesnt exist, unique variable of 1st spectrum
            # if it fails, let it fail
    # Make sure it is in both Spectra
    if var not in s1.get_vars():
        raise KeyError("Variable {0} not in Spectrum {1}".format(var, s1.get_name()))
    if var not in s2.get_vars():
        raise KeyError("Variable {0} not in Spectrum {1}".format(var, s1.get_name()))

    # Use same units
    Iunit1 = s1.units[var]
    wunit1 = s1.get_waveunit()

    # Resample s2 on s1
    s2 = s2.resample(s1, inplace=False)

    # Subtract
    w1, I1 = s1.get(var=var, Iunit=Iunit1, wunit=wunit1)
    w2, I2 = s2.get(var=var, Iunit=Iunit1, wunit=wunit1)

    name = s1.get_name() + "-" + s2.get_name()

    sub = Spectrum.from_array(w1, I1 - I2, var, wunit=wunit1, Iunit=Iunit1, name=name)
    #    warn("Conditions of the left spectrum were copied in the substraction.", Warning)
    return sub


def concat_spectra(s1, s2, var=None):
    """Concatenate two spectra ``s1`` and ``s2`` side by side.

    Note: their spectral range should not overlap

    Returns
    -------

    s: Spectrum
        Spectrum object with the same units and waveunits as ``s1``

    Parameters
    ----------

    s1, s2: Spectrum objects
        Spectrum you want to concatenate
    var: str
        quantity to manipulate: 'radiance', 'transmittance', ... If ``None``,
        get the unique spectral quantity of ``s1``, or the unique spectral
        quantity of ``s2``, or raises an error if there is any ambiguity

    Notes
    -----

    .. warning::

        the output Spectrum has the sum of the spectral ranges of s1 and s2.
        It won't be evenly spaced. This means that you cannot apply a slit without
        side effects. Typically, you want to use this function for convolved
        quantities only, such as experimental spectra. Else, use
        :func:`~radis.los.slabs.MergeSlabs` with the options
        ``resample='full', out='transparent'``

    See Also
    --------

    :func:`~radis.spectrum.operations.add_spectra`,
    :func:`~radis.los.slabs.MergeSlabs`

    """

    # Get variable
    if var is None:
        try:
            var = _get_unique_var(
                s2, var, inplace=False
            )  # unique variable of 2nd spectrum
        except KeyError:
            var = _get_unique_var(
                s1, var, inplace=False
            )  # if doesnt exist, unique variable of 1st spectrum
            # if it fails, let it fail
    # Make sure it is in both Spectra
    if var not in s1.get_vars():
        raise KeyError("Variable {0} not in Spectrum {1}".format(var, s1.get_name()))
    if var not in s2.get_vars():
        raise KeyError("Variable {0} not in Spectrum {1}".format(var, s1.get_name()))

    if var in ["transmittance_noslit", "transmittance"]:
        warn(
            "It does not make much physical sense to sum transmittances. Are "
            + "you sure of what you are doing? See also // (MergeSlabs) and > "
            + "(SerialSlabs)"
        )

    # Use same units
    Iunit1 = s1.units[var]
    wunit1 = s1.get_waveunit()

    # Get the value, on the same wunit)
    w1, I1 = s1.get(var=var, copy=False)  # @dev: faster to just get the stored value.
    # it's copied in hstack() below anyway).
    w2, I2 = s2.get(var=var, Iunit=Iunit1, wunit=wunit1)

    if not (w1.max() < w2.min() or w2.max() > w1.min()):
        raise ValueError(
            "You cannot use concat_spectra for overlapping spectral ranges. "
            + "Got: {0:.2f}-{1:.2f} and {2:.2f}-{3:.2f} {4}. ".format(
                w1.min(), w1.max(), w2.min(), w2.max(), wunit1
            )
            + "Use MergeSlabs instead, with the correct `out=` parameter "
            + "for your case"
        )

    w_tot = hstack((w1, w2))
    I_tot = hstack((I1, I2))

    name = s1.get_name() + "&" + s2.get_name()  # use "&" instead of "+"

    concat = Spectrum.from_array(
        w_tot, I_tot, var, wunit=wunit1, unit=Iunit1, name=name
    )

    return concat


def offset(s, offset, unit, name=None, inplace=False):
    # type: (Spectrum, float, str, str, bool) -> Spectrum
    """Offset the spectrum by a wavelength or wavenumber

    Parameters
    ----------
    s: Spectrum
        Spectrum you want to modify
    offset: float
        Constant to add to all quantities in the Spectrum.
    unit: 'nm' or 'cm-1'
        unit for ``offset``.

    Other Parameters
    ----------------
    name: str
        name of output spectrum
    inplace: bool
        if ``True``, modifies ``s`` directly. Else, returns a copy.
        Default ``False``

    Returns
    -------
    s : Spectrum
        Spectrum object where cst is added to intensity of s['var']
        If ``inplace=True``, ``s`` has been modified directly.

    See Also
    --------
    call as a Spectrum method directly: :py:meth:`~radis.spectrum.spectrum.Spectrum.offset`
    """

    stored_waveunit = s.get_waveunit()

    # Convert offset to correct unit:
    if stored_waveunit == "cm-1":
        if unit == "nm":
            # Note @EP: here we're offsetting by a constant value in 'nm', which is
            # not a constant value in 'cm-1'. The offset is an array
            offset_q = -dnm_air2dcm(offset, s.get_wavelength())  # this is an array
        elif unit == "nm_vac":
            offset_q = -dnm2dcm(offset, s.get_wavelength())  # this is an array
        elif unit == "cm-1":
            offset_q = offset
        else:
            raise ValueError(unit)
    elif stored_waveunit == "nm":  # wavelength air
        if unit == "nm":
            offset_q = offset
        elif unit == "nm_vac":
            # Note @EP: strictly speaking, the offset should change a little bit
            # as nm_vac > nm depends on the wavelength. Neglected here. # TODO ?
            offset_q = offset
        elif unit == "cm-1":
            offset_q = -dcm2dnm_air(offset, s.get_wavenumber())  # this is an array
        else:
            raise ValueError(unit)
    elif stored_waveunit == "nm_vac":  # wavelength vacuum
        if unit == "nm":
            # Note @EP: strictly speaking, the offset should change a little bit
            # as nm > nm_vac depends on the wavelength. Neglected here. # TODO ?
            offset_q = offset
        elif unit == "nm_vac":
            offset_q = offset
        elif unit == "cm-1":
            offset_q = -dcm2dnm(offset, s.get_wavenumber())  # this is an array
        else:
            raise ValueError(unit)
    else:
        raise ValueError(stored_waveunit)

    if not inplace:
        s = s.copy()

    # Update all variables
    s._q["wavespace"] += offset_q
    # @dev: updates the Spectrum directly because of copy=False

    if name:
        s.name = name

    return s


def get_baseline(
    s, var="radiance", algorithm="als", wunit="default", Iunit="default", **kwargs
):
    """Calculate and returns a baseline

    Parameters
    ----------
    s: Spectrum
        Spectrum which needs a baseline
    var: str
        on which spectral quantity to read the baseline. Default ``'radiance'``.
        See :py:data:`~radis.spectrum.utils.SPECTRAL_QUANTITIES`
    algorithm: 'als', 'polynomial'
        Asymmetric least square or Polynomial fit
    **kwargs: dict
        additional parameters to send to the algorithm. By default,
        for 'polynomial':

            **kwargs = **{"deg": 1, "max_it":500}

        for 'als':

            **kwargs = {"asymmetry_param": 0.05,
                        "smoothness_param": 1e6}

    Returns
    -------
    baseline: Spectrum
        Spectrum object where intensity is the baseline of s is computed by peakutils

    Examples
    --------

    .. minigallery:: radis.Spectrum.get_baseline

    See Also
    --------

    :py:func:`~radis.spectrum.operations.sub_baseline`,
    :py:meth:`~radis.spectrum.spectrum.Spectrum.get_baseline`
    """
    if wunit == "default":
        wunit = s.get_waveunit()
    if Iunit == "default":
        Iunit = s.units[var]

    w1, I1 = s.get(var=var, wunit=wunit, Iunit=Iunit)

    if algorithm == "polynomial":
        from radis.misc.signal import baseline

        polyargs = {"deg": 1, "max_it": 500}
        polyargs.update(kwargs)
        baseline = baseline(I1, **polyargs)
    elif algorithm == "als":
        from radis.misc.signal import als_baseline

        alsargs = {"asymmetry_param": 0.05, "smoothness_param": 1e6}
        alsargs.update(kwargs)
        baseline = als_baseline(I1, **alsargs)

    baselineSpectrum = Spectrum.from_array(
        w1, baseline, var, wunit=wunit, Iunit=Iunit, name=s.get_name() + "_baseline"
    )
    return baselineSpectrum


# %% Tests

if __name__ == "__main__":
    from radis.test.spectrum.test_operations import _run_testcases

    _run_testcases(plot=True)



# import sys
from ctypes import (
    byref,
    c_char_p,
    c_float,
    c_int,
    c_long,
    c_size_t,
    c_void_p,
    cast,
    create_string_buffer,
)
from os import name as os_name
from warnings import warn

if os_name == "nt":
    from ctypes import windll as dllobj
else:
    from ctypes import cdll as dllobj

import numpy as np

from radis.misc.utils import NotInstalled, not_installed_nvidia_args

try:
    from nvidia.cufft import __path__ as cufft_path
except ImportError:
    cufft_path = NotInstalled(*not_installed_nvidia_args)

_verbose = False


def getClasses():
    return CuContext, CuModule, CuArray, CuFFT, CuTimer


def getCUDAVersion(ptx_file):
    # Reads the version of the CUDA Toolkit that was used to compile PTX file
    with open(ptx_file, "rb") as f:
        for i in range(5):
            line = f.readline()
        version_string = line.decode().split(",")[2]
        version_string = version_string[version_string.find("V") + 1 :]
        major, minor, patch = map(int, version_string.split("."))
        return major, minor, patch


def getRequiredDriverVersion(ptx_file):
    # Returns the minimal required CUDA driver version
    major, minor, patch = getCUDAVersion(ptx_file)
    if major == 11:
        if minor == 0:
            drv_ver = "v451.22" if os_name == "nt" else "v450.36.06"
        else:
            drv_ver = "v452.39" if os_name == "nt" else "v450.80.02"

    elif major == 12:
        drv_ver = "v527.41" if os_name == "nt" else "v525.60.13"

    else:
        drv_ver = "[latest version]"

    return drv_ver


def cu_print(*vargs):
    global _verbose
    if _verbose:
        print(*vargs)


CUDA_SUCCESS = 0
CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK = 1
CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR = 75
CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR = 76

CUFFT_R2C = 0x2A
CUFFT_C2R = 0x2C

lib = None
lib_cufft = None

_arrays = []
_plans = []
_modules = []


class CuContext:
    def __init__(self, device, context, stream):

        # private:
        self._device = device
        self._context = context
        self._stream = stream

    @staticmethod
    def setVerbosity(level):
        global _verbose
        if level >= 2:
            _verbose = level

    @staticmethod
    def Open(device_id=0, flags=0, verbose=None):
        global lib

        if verbose is not None:
            CuContext.setVerbosity(verbose)

        # private:
        _context = c_void_p(0)
        _device = c_void_p(0)
        _stream = c_void_p(0)

        cuda_name = "nvcuda.dll" if os_name == "nt" else "libcuda.so"
        try:
            lib = dllobj.LoadLibrary(cuda_name)

        except Exception:
            if os_name == "nt":
                print("Can't find {:s}...".format(cuda_name))
                return None

            else:
                cuda_name = "libcuda.so.1"
                try:
                    lib = dllobj.LoadLibrary(cuda_name)
                except (OSError):
                    print("Can't find {:s}...".format(cuda_name))
                    return None

        err = lib.cuInit(0)

        deviceCount = c_int(0)
        lib.cuDeviceGetCount(byref(deviceCount))

        if deviceCount == 0:
            print("Error: no devices supporting CUDA\n")
            return None

        lib.cuDeviceGet(byref(_device), device_id)

        err = lib.cuCtxCreate_v2(byref(_context), flags, _device)
        if err != CUDA_SUCCESS:
            print("Error initializing the CUDA context.")
            lib.cuCtxDestroy_v2(_context)
            return None

        # cu_print(lib.cuStreamCreate(byref(_stream), 0), "ctx.create stream")

        return CuContext(_device, _context, _stream)

    @staticmethod
    def getDeviceList():
        dev_list = []
        _deviceCount = c_int(0)

        lib.cuInit(0)
        lib.cuDeviceGetCount(byref(_deviceCount))

        for i in range(_deviceCount.value):

            _device = c_long(0)
            lib.cuDeviceGet(byref(_device), i)

            _name = create_string_buffer(100)
            lib.cuDeviceGetName(_name, 100, _device)
            dev_list.append(_name.value.decode())

        return dev_list

    def printDeviceCapabilities(self):

        _major = c_long(0)
        _minor = c_long(0)
        lib.cuDeviceGetAttribute(
            byref(_major), CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MAJOR, self._device
        )
        lib.cuDeviceGetAttribute(
            byref(_minor), CU_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY_MINOR, self._device
        )
        print(
            "> GPU Device has SM {:d}.{:d} compute capability".format(
                _major.value, _minor.value
            )
        )

        _totalGlobalMem = c_size_t(0)
        lib.cuDeviceTotalMem_v2(byref(_totalGlobalMem), self._device)
        print(
            "  Total amount of global memory:   {:d} bytes".format(
                _totalGlobalMem.value
            )
        )
        print(
            "  64-bit Memory Address:           {:s}".format(
                "YES" if _totalGlobalMem.value > (2 << 31) else "NO"
            )
        )

    def getMaxThreadsPerBlock(self):
        _Ntpb = c_long(0)
        lib.cuDeviceGetAttribute(
            byref(_Ntpb), CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK, self._device
        )
        return _Ntpb.value

    def setCurrent(self):
        cu_print(lib.cuCtxSetCurrent(self._context), "ctx.setcurrent")

    def synchronize(self):
        cu_print(lib.cuCtxSynchronize(), "ctx.sync")

    def syncStream(self):
        cu_print(lib.cuStreamSynchronize(0), "ctx.sync stream")  # _ptsz

    def destroy(self):
        global _arrays, _plans, _modules

        while len(_arrays):
            arr = _arrays.pop(0)
            arr.free()

        while len(_plans):
            plan = _plans.pop(0)
            plan.destroy()

        while len(_modules):
            mod = _modules.pop(0)
            mod.unload()

        if self._stream.value is not None:
            cu_print(lib.cuStreamDestroy_v2(self._stream), "ctx.destroy stream")

        try:
            cu_print(lib.cuCtxDestroy_v2(self._context), "ctx.destroy")
        except (AttributeError):
            pass


class CuModule:
    def __init__(self, context, module_name):

        # public:
        self.module_name = module_name
        self.context = context
        self.mode = "GPU"
        global _modules
        _modules.append(self)

        # private:
        self._module = c_void_p(0)
        self._func_dict = {}
        self._global_dict = {}

        _module_name = c_char_p(self.module_name.encode())
        err = lib.cuModuleLoad(byref(self._module), _module_name)
        if err != CUDA_SUCCESS:
            if err == 222:
                drv_ver = getRequiredDriverVersion(module_name)
                print(
                    "\n\n*** CUDA Driver too old!***\nMinimally required driver version is {:s}\n".format(
                        drv_ver
                    )
                    + "Please update driver by downloading it at:\n-> www.nvidia.com/download/index.aspx \n"
                )

            warn(
                "* Error loading the module {:s}\n".format(_module_name.value.decode())
            )
            self.context.destroy()
            return

    def __getattr__(self, attr):
        try:
            self._func_dict[attr]
            return self._func_dict[attr]

        except (KeyError):
            _function = c_void_p(0)
            _kernel_name = c_char_p(attr.encode())

            err = lib.cuModuleGetFunction(byref(_function), self._module, _kernel_name)
            if err != CUDA_SUCCESS:
                print(
                    "* Error getting kernel function {:s}".format(
                        _kernel_name.value.decode()
                    )
                )
                self.context.destroy()
                # sys.exit()

            self._func_dict[attr] = CuFunction(_function)
            self._func_dict[attr].module = self
            return self._func_dict[attr]

    def _getGlobal(self, name, ctype=None):
        try:
            return self._global_dict[name]

        except (KeyError):
            if ctype is not None:
                _var = c_void_p()
                _size = c_size_t()
                _type = ctype
                cu_print(
                    lib.cuModuleGetGlobal_v2(
                        byref(_var), byref(_size), self._module, c_char_p(name.encode())
                    ),
                    "mod.global",
                )
                self._global_dict[name] = (_var, _size, _type)
                cu_print("size: ", _size.value)
                return self._global_dict[name]

            else:
                print("Constant type must be passed first time it is called!")
                return

    def getMode(self):
        return self.mode

    def setConstant(self, name, cval):
        _var, _size, _type = self._getGlobal(name, type(cval))
        cu_print(lib.cuMemcpyHtoD_v2(_var, byref(cval), _size), "mod.HtoD")

    def getConstant(self, name, ctype=None):
        _var, _size, _type = self._getGlobal(name, ctype)
        cvar = _type()
        cu_print(lib.cuMemcpyDtoH_v2(byref(cvar), _var, _size), "mod.DtoH")
        return cvar

    def unload(self):
        cu_print(lib.cuModuleUnload(self._module), "mod.unload")


class CuFunction:
    def __init__(self, _function):

        # public:
        self.module = None
        self.args = []
        self.blocks = (1, 1, 1)
        self.threads = (1, 1, 1)
        self.sync = False

        # private:
        self._function = _function

    def setGrid(self, blocks=(1, 1, 1), threads=(1, 1, 1)):
        self.blocks = blocks
        self.threads = threads

    def setArgs(self, *vargs):
        self.args = vargs

    def __call__(self, *vargs, blocks=None, threads=None, sync=None):

        if len(vargs):
            self.args = vargs
        if blocks is not None:
            self.blocks = blocks
        if threads is not None:
            self.threads = threads
        if sync is not None:
            self.sync = sync

        _stream = c_void_p(0)  # default stream; else: self.module.context._stream
        voidPtrArr = len(self.args) * c_void_p
        _c_args = voidPtrArr(*[cast(byref(arr._ptr), c_void_p) for arr in self.args])
        _extra_args = (1 * c_void_p)(0)

        cu_print(
            lib.cuLaunchKernel(  # _ptsz
                self._function,
                *self.blocks,
                *self.threads,
                0,  # shared memory size #TODO: get capability; set max size
                _stream,
                _c_args,
                _extra_args,
            ),
            "func.kernel",
        )

        if self.sync:
            self.module.context.synchronize()


class CuArray:
    def __init__(self, shape, dtype=np.float32, init="empty", grow_only=False):
        self._ptr = c_void_p()
        self._nbytes_alloc = 0
        self.grow_only = grow_only
        self.resize(shape, dtype, init)
        global _arrays
        _arrays.append(self)

    def resize(self, shape=None, dtype=None, init="empty"):

        shape_tuple = shape if isinstance(shape, tuple) else (shape,)
        self.shape = self.shape if shape is None else shape_tuple
        self.dtype = self.dtype if dtype is None else np.dtype(dtype)
        self.size = int(np.prod(self.shape))
        self.itemsize = self.dtype.itemsize
        self.nbytes = self.itemsize * self.size

        if init not in ("zeros", "empty"):
            return

        if self.nbytes > self._nbytes_alloc or (
            (self.nbytes < self._nbytes_alloc) and (not self.grow_only)
        ):
            if self._ptr.value is not None:
                self.free()
                self._ptr = c_void_p(0)

            cu_print(lib.cuMemAlloc_v2(byref(self._ptr), self.nbytes), "arr.alloc")
            self._nbytes_alloc = self.nbytes

            if self._ptr.value is not None:
                cu_print(hex(self._ptr.value), self.nbytes)

        if init == "zeros":
            self.zeroFill()

    @staticmethod
    def fromArray(arr):

        obj = CuArray(arr.shape, arr.dtype, init=None)
        obj.setArray(arr)

        return obj

    def zeroFill(self):
        cu_print(lib.cuMemsetD8_v2(self._ptr, 0, self.nbytes), "arr.zeros")

    def setArray(self, arr):

        params_changed = arr.shape != self.shape or arr.dtype != self.dtype
        uninitialized_memory = self._ptr.value is None
        if params_changed or uninitialized_memory:
            self.resize(arr.shape, arr.dtype, "empty")

        cu_print(
            lib.cuMemcpyHtoD_v2(self._ptr, c_void_p(arr.ctypes.data), self.nbytes),
            "arr.HtoD",
        )

    def getArray(self):

        arr = np.empty(self.shape, dtype=self.dtype)
        cu_print(
            lib.cuMemcpyDtoH_v2(c_void_p(arr.ctypes.data), self._ptr, self.nbytes),
            "arr.DtoH",
        )
        return arr

    def free(self):
        cu_print(lib.cuMemFree_v2(self._ptr), "arr.free")


class CuFFT:
    def __init__(self, arr_in, arr_out, workarea=None, direction="fwd"):
        global lib_cufft, _plans
        _plans.append(self)

        # public:
        self.arr_in = arr_in
        self.arr_out = arr_out
        self.workarea = (
            CuArray(0, dtype=np.byte, grow_only=True) if workarea is None else workarea
        )

        # private:
        self._direction = direction
        self._fft_type = CUFFT_R2C if direction == "fwd" else CUFFT_C2R
        self._arr = arr_in if direction == "fwd" else arr_out
        self._plans = {}

        cufft_name = (
            "\\bin\\cufft64_10.dll" if os_name == "nt" else "/lib/libcufft.so.10"
        )
        try:
            if lib_cufft is None:
                lib_cufft = dllobj.LoadLibrary(cufft_path[0] + cufft_name)
        except (FileNotFoundError):
            print("Can't find {:s}...".format(cufft_name))
            return None

    def _getPlan(self):
        try:
            return self._plans[self._arr.shape]

        except (KeyError):

            _plan = c_void_p(0)
            cu_print(lib_cufft.cufftCreate(byref(_plan)), "fft.plan create")
            cu_print(lib_cufft.cufftSetAutoAllocation(_plan, c_int(0)), "fft.setalloc")

            batch = int(np.prod(self._arr.shape[1:]))
            oneInt = 1 * c_int
            _n = oneInt(self._arr.shape[0])
            stride = batch
            oneSizeT = 1 * c_size_t
            _worksize = oneSizeT(0)

            cu_print(
                lib_cufft.cufftMakePlanMany(
                    _plan,
                    1,  # rank,
                    _n,  # n
                    oneInt(0),  # inembed,
                    stride,  # istride
                    1,  # idist,
                    oneInt(0),  # onembed,
                    stride,  # ostride
                    1,  # odist,
                    self._fft_type,
                    batch,
                    _worksize,
                ),
                "fft.plan many",
            )

            self.workarea.resize((_worksize[0],), dtype=np.byte)
            cu_print(
                lib_cufft.cufftSetWorkArea(_plan, self.workarea._ptr),
                "fft.set workarea",
            )

            self._plans[self._arr.shape] = [_plan, int(self.workarea._ptr.value)]
            return self._plans[self._arr.shape]

    def __call__(self):

        _plan, ptrval = self._getPlan()

        if ptrval != self.workarea._ptr.value:
            cu_print(
                lib_cufft.cufftSetWorkArea(_plan, self.workarea._ptr),
                "fft.set workarea",
            )
            self._plans[self._arr.shape][1] = int(self.workarea._ptr.value)

        if self._direction == "fwd":
            cu_print(
                lib_cufft.cufftExecR2C(_plan, self.arr_in._ptr, self.arr_out._ptr),
                "fft.fwd",
            )
        else:
            cu_print(
                lib_cufft.cufftExecC2R(_plan, self.arr_in._ptr, self.arr_out._ptr),
                "fft.rev",
            )

    def destroy(self):

        for key in self._plans:
            _plan, workarea = self._plans[key]
            cu_print(lib_cufft.cufftDestroy(_plan), "fft.destroy")


class CuTimer:
    def __init__(self, stream=0, flags=0):
        # private:
        self._stream = c_int(stream)
        self._start = c_void_p(0)
        self._stop = c_void_p(0)

        # public:
        self.times = {}

        cu_print(lib.cuEventCreate(byref(self._start), flags), "time.create")
        cu_print(lib.cuEventCreate(byref(self._stop), flags), "time.create")
        self.reset()

    def reset(self):
        cu_print(lib.cuEventRecord(self._start, self._stream), "time.record")  # _ptsz

    def lap(self, name=None):
        if name is None:
            name = "event{:d}".format(len(self.times.keys()))
        self.times[name] = self()

    def __call__(self):
        _time = c_float(0)  # elapsed time in ms
        cu_print(lib.cuEventRecord(self._stop, self._stream), "time.record")  # _ptsz
        cu_print(lib.cuEventSynchronize(self._stop), "time.sync")
        cu_print(
            lib.cuEventElapsedTime(byref(_time), self._start, self._stop), "time.time"
        )
        return _time.value

    def getTimes(self):
        return self.times

    def getDiffs(self):
        vals = [*self.times.values()]
        diffs = [vals[0]] + [vals[i] - vals[i - 1] for i in range(1, len(vals))]
        return dict(zip(self.times.keys(), diffs))



# -*- coding: utf-8 -*-
"""
Comparing an atomic spectrum for the Kurucz database with that generated from SpectraPlot (https://spectraplot.com/)
1. `test_Kurucz_vs_NISTandSpectraplot` uses `SpectrumFactory` directly, works in nm, tests the absorption coefficient, and uses Pandas
2. `test_Kurucz_vs_NISTandSpectraplot_4000` uses `calc_spectrum`, with values for most parameters largely equivalent to the defaults in SpectraPlot, works in cm-1, tests the integral under the emission graphs, uses Vaex, and produces verbose output
Note that SpectraPlot appears to use the 'Observed Wavelength Air' given by NIST, whereas the wavelengths in the Kurucz database seem to equate to the more precise 'Ritz Wavelength Air' in NIST, so there may be a relative shift between the spectrum from RADIS and that from SpectraPlot
"""
import numpy as np
import pytest

import radis
from radis import Spectrum, SpectrumFactory, calc_spectrum, plot_diff
from radis.misc.utils import NotInstalled, not_installed_vaex_args
from radis.test.utils import getValidationCase

try:
    import vaex
except ImportError:
    vaex = NotInstalled(*not_installed_vaex_args)


def func1(**kwargs):
    """An example implementing the default broadening formula and values of SpectraPlot"""
    # print(kwargs.keys())
    # print(kwargs['df'].columns)
    return 0.1 * (296 / kwargs["Tgas"]) ** 0.8, None


@pytest.mark.needs_connection
def test_Kurucz_vs_NISTandSpectraplot(plot=True, verbose=True):
    def broad_arbitrary(**kwargs):
        """An arbitrary broadening formula in SpectraPlot (https://spectraplot.com/)"""
        return 1 * (296 / kwargs["Tgas"]) ** 0.8, None

    #%% Employ the same inputs than in example file 'spectraplot_O_10000K.txt'
    sf = SpectrumFactory(
        wavelength_min=777,
        wavelength_max=778,
        wstep=0.001,
        species="O_I",
        optimization="simple",
        path_length=1,  # cm
        pressure=1,  # atm
        verbose=0,
        lbfunc=broad_arbitrary,
    )
    sf.fetch_databank("kurucz", parfuncfmt="kurucz")
    # sf.load_databank('Kurucz-O_I', drop_columns=[], load_columns='all')
    s_RADIS = sf.eq_spectrum(Tgas=10000, name="Kurucz by RADIS")
    # s_RADIS.plot("radiance_noslit", wunit="cm-1")

    #%% Experimental spectrum
    L = 1  # cm - Input in SpectraPlot software
    raw_data = np.loadtxt(
        getValidationCase("spectraplot_O_10000K.txt"), delimiter=",", skiprows=1
    )
    s_SpectraPlot = Spectrum(
        {
            "wavenumber": raw_data[:, 0],
            "abscoeff": raw_data[:, 1]
            / L,  # spectraplot outputs absorbance; abscoef = absorbance / L
        },
        units={"abscoeff": "cm-1"},
        wunit="cm-1",
        name="NIST by SpectraPlot",
    )
    # s_SpectraPlot.plot('abscoeff', wunit='cm-1')

    if plot:
        plot_diff(s_RADIS, s_SpectraPlot, "abscoeff", wunit="nm")
    A_RADIS = s_RADIS.get_integral("abscoeff", wunit="nm")
    A_SpectraPlot = s_SpectraPlot.get_integral("abscoeff", wunit="nm")

    if verbose:
        print(
            f"Ratio of area under abscoef ('k') is A_RADIS/A_SpectraPlot = {A_RADIS/A_SpectraPlot:.3f}"
        )

    assert np.isclose(A_RADIS, A_SpectraPlot, rtol=1e-2)


@pytest.mark.needs_connection
@pytest.mark.skipif(isinstance(vaex, NotInstalled), reason="Vaex not available")
def test_Kurucz_vs_NISTandSpectraplot_4000(plot=True, verbose=True):
    initial_engine = radis.config["DATAFRAME_ENGINE"]
    radis.config["DATAFRAME_ENGINE"] = "vaex"

    w, I = np.loadtxt(
        getValidationCase("spectraplot_O_4000K.csv"),
        skiprows=1,
        delimiter=",",
        unpack=True,
    )
    I = np.where(I == 0, 1e-99, I)

    s_SpectraPlot = Spectrum.from_array(
        w, I, quantity="radiance_noslit", wunit="cm-1", Iunit="μW/cm2/sr/cm-1"
    )

    s_RADIS = calc_spectrum(
        12850,
        12870,  # from this up to 13120 is largely 0
        species="O",  # should be converted to O_I by radis.db.classes.to_conventional_name
        Tgas=4000,  # K
        databank="kurucz",
        pressure=1.01325,
        path_length=15,
        lbfunc=func1,
        warnings={"AccuracyError": "ignore", "AccuracyWarning": "ignore"},
        verbose=2,
    )

    # s_RADIS.plot("radiance_noslit", wunit="cm-1")
    if plot:
        plot_diff(
            s_SpectraPlot, s_RADIS, label1="SpectraPlot", label2="Kurucz"
        )  # , method='ratio')

    I_RADIS = s_RADIS.get_integral(
        "radiance_noslit", wunit="cm-1", Iunit="mW/cm2/sr/cm-1"
    )  # , return_units=True)
    I_SpectraPlot = s_SpectraPlot.get_integral(
        "radiance_noslit", wunit="cm-1", Iunit="mW/cm2/sr/cm-1"
    )  # , return_units=True)
    # print(I_RADIS, I_SpectraPlot, I_RADIS - I_SpectraPlot, (I_RADIS - I_SpectraPlot)/I_SpectraPlot)
    if verbose:
        print(
            f"Ratio of area under emission (radiance) is I_RADIS/I_SpectraPlot = {I_RADIS/I_SpectraPlot}"
        )

    assert np.isclose(I_RADIS, I_SpectraPlot, rtol=1.4e-2)

    radis.config["DATAFRAME_ENGINE"] = initial_engine


if __name__ == "__main__":
    test_Kurucz_vs_NISTandSpectraplot()
    test_Kurucz_vs_NISTandSpectraplot_4000()



# -*- coding: utf-8 -*-
"""
Test spectrum creation / export
"""

import pytest

from radis import Spectrum, get_residual, test_spectrum


@pytest.mark.fast
def test_specutils_io(verbose=True, plot=False, *args, **kwargs):
    """Test Radis/specutils  interface in particular when dealing with wavelengths

    see https://github.com/radis/radis/pull/499
    """

    s = test_spectrum().take("transmittance_noslit")

    processed_spec = Spectrum.from_array(
        *s.get("transmittance_noslit", wunit="nm"),
        "transmittance_noslit",
        wunit="nm",
        Iunit="",
    )

    with pytest.raises(ValueError) as err:

        spec_u = processed_spec.to_specutils(
            "transmittance_noslit", wunit="nm_air", Iunit="default"
        )
        assert (
            "specutil's Spectrum1D will only handle wavelengths as seen in vacuum"
            in str(err)
        )

    spec_u = processed_spec.to_specutils(
        "transmittance_noslit", wunit="nm_vac", Iunit="default"
    )

    if plot:

        import matplotlib.pyplot as plt

        s.plot("transmittance_noslit", wunit="nm_vac", lw=5)
        plt.plot(spec_u.wavelength.to("nm"), spec_u.flux, lw=3)
        plt.xlim((4591.91, 4597.57))

    assert spec_u.meta["waveunit"] == "nm_vac"  # was corrected properly
    # Reverse

    s2 = Spectrum.from_specutils(spec_u, "transmittance_noslit")

    if plot:

        s2.plot("transmittance_noslit", wunit="nm_vac", lw=1, nfig="same")

    assert get_residual(s, s2, "transmittance_noslit", ignore_nan=True) < 1e-9


@pytest.mark.fast
def test_reading_from_Matlab(verbose=True, plot=False, *args, **kwargs):
    """Test creating a Spectrum from a Matlab file with
    :py:meth:`~radis.spectrum.spectrum.Spectrum.from_mat`
    """

    from radis import Spectrum
    from radis.test.utils import getTestFile

    s = Spectrum.from_mat(
        getTestFile("trimmed_1857_VoigtCO_Minesi.mat"),
        "absorbance",
        wunit="cm-1",
        unit="",
        index=10,
    )

    import numpy as np

    assert np.isclose(s.get("absorbance")[0][100], 2011.5356487633755)
    assert np.isclose(s.get("absorbance")[1][100], 0.005560075444710532)


if __name__ == "__main__":
    test_specutils_io(plot=True)
    test_reading_from_Matlab()



# -*- coding: utf-8 -*-
"""
Test that :class:`~radis.tools.database.SpecDatabase` works

-------------------------------------------------------------------------------

"""


import os
from os.path import dirname, exists

import pytest

from radis.test.utils import getTestFile
from radis.tools.database import SpecDatabase, SpecList


@pytest.mark.fast
def test_speclist(*args, **kwargs):
    """Test :py:class:`~radis.tools.database.SpecList` functions"""

    from radis.test.utils import getTestFile

    # Get a list of test files
    spec_list = SpecDatabase(getTestFile("."), lazy_loading=False).get()

    # Now generate a SpecList
    dbl = SpecList(*spec_list)
    # check methods work:
    s_closest = dbl.get_closest(Trot=1550)  # TODO add which one
    assert s_closest.c["Trot"] == 1500

    # test get_closest() with a non numeric argument
    s_closest_2 = dbl.get_closest(Trot=1550, molecule="CO")
    assert s_closest_2.c["Trot"] == 1500


@pytest.mark.fast
def test_database_functions(
    verbose=True, plot=True, close_plots=True, warnings=True, *args, **kwargs
):
    """Test :py:class:`~radis.tools.database.SpecDatabase` functions"""

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()
    if close_plots:
        plt.close("all")

    db = SpecDatabase(dirname(getTestFile(".")), lazy_loading=False)

    # Database visualisation methods
    if verbose:
        print(
            "{0} items in test database: {1}".format(len(db), db.see(["Tvib", "Trot"]))
        )
    if plot:
        db.plot_cond("Tvib", "Trot")

    # Database get methods
    db.get_closest(Tgas=1300, path_length=1)
    s = db.get_unique(
        Tgas=1500, path_length=0.01, mole_fraction=0.5
    )  # there should be one only
    # ... note that this is just so we can test
    # ... get_unique(). If we were to add new
    # ... test cases with matching conditions
    # ... let's add more criteria to keep it unique
    match = db.get(**s.conditions)
    assert len(match) == 1

    # Database add method
    s2 = s.copy()
    s2.conditions["Tgas"] = 0  # make it unique (for testing)

    matching_spectrum_in_db = s2 in db

    l = db.add(
        s2,
        add_info=["Tvib", "Trot"],
        discard=[],
        compress=True,
        if_exists_then="increment",
    )
    assert exists(l)
    try:
        assert s2 in db
        # .. ensures that you cant add it twice
        with pytest.raises(ValueError):
            db.add(s2, add_info=["Tvib", "Trot"], if_exists_then="error")

    # Now remove the Spectrum, update the database and ensures it's not there anymore
    finally:
        os.remove(l)
    db.update(force_reload=True)  # update database
    assert not exists(l)

    # make sure we deleted it properly
    if not matching_spectrum_in_db:
        assert s2 not in db


def test_plot_spec(plot=True, close_plots=True, verbose=True, *args, **kwargs):

    from radis.test.utils import getTestFile
    from radis.tools.database import plot_spec

    if plot:
        import matplotlib.pyplot as plt

        plt.ion()
    if close_plots:
        plt.close("all")

    if plot:
        plot_spec(getTestFile("N2C_specair_380nm.spec"))


def test_save_compressed2(verbose=True, *args, **kwargs):
    "Check if saving a spectrum with compress = 2 does not change something."
    import shutil
    from os.path import join

    from radis import SpecDatabase, calc_spectrum
    from radis.test.utils import setup_test_line_databases

    shutil.rmtree(join(dirname(getTestFile(".")), "newDb"), ignore_errors=True)

    # get the spectrum
    setup_test_line_databases()
    s = calc_spectrum(
        2000,
        2300,  # cm-1
        molecule="CO",
        isotope="1,2,3",
        pressure=1.01325,  # bar
        Tgas=700,  # K
        mole_fraction=0.1,
        path_length=1,  # cm
        verbose=False,
        wstep=0.01,
        medium="vacuum",
        databank="HITRAN-CO-TEST",
    )
    try:

        # load in one database
        db = SpecDatabase(join(dirname(getTestFile(".")), "newDb"))
        db.add(s, compress=2, if_exists_then="error")

        # simulate an experimentalist who come later and load the spectrum
        db2 = SpecDatabase(join(dirname(getTestFile(".")), "newDb"))
        s_bis = db2.get_unique(Tgas=700)

    finally:
        # we want to make sure this folder is deleted
        shutil.rmtree(join(dirname(getTestFile(".")), "newDb"))

    # we check the loaded spectrum contains less information than the calculated one
    assert not s == s_bis
    assert s_bis.get_vars() == ["abscoeff"]  # only this spectral quantity was stored
    assert s_bis.lines is None
    assert s_bis.conditions is not None  # we kept the metadata
    # now we check if it works
    s_bis.update()
    for var in s.get_vars():
        assert s.compare_with(s_bis, spectra_only=var, plot=False, verbose=verbose)


def test_lazy_loading(verbose=True, *args, **kwargs):
    """Test lazy loading"""

    import numpy as np

    from radis import calc_spectrum

    DBNAME = "test_CO_OH_database"
    db = SpecDatabase(DBNAME, lazy_loading=False)  # create the database

    #%% Simulate a fake database

    # Compute spectra of CO and OH and add them to the database
    T_db = np.linspace(300, 1200, 5)  # Temperature array
    for T in T_db:
        s1 = (
            calc_spectrum(
                wavenum_min=3000,
                wavenum_max=4500,
                molecule="CO",
                Tgas=T,
                databank="hitemp",  # test by fetching directly
                verbose=False,
            )
            .apply_slit(2, "nm")
            .take("radiance")
            # TODO : use https://github.com/radis/radis/issues/135 once implemented
        )
        db.add(s1, store_name=f"CO_{T}K.spec", if_exists_then="ignore")
        s2 = (
            calc_spectrum(
                wavenum_min=3000,
                wavenum_max=4500,
                molecule="OH",
                Tgas=T,
                databank="hitemp",  # test by fetching directly
                verbose=False,
            )
            .apply_slit(2, "nm")
            .take("radiance")
            # TODO : use https://github.com/radis/radis/issues/135 once implemented
        )
        db.add(s2, store_name=f"OH_{T}K.spec", if_exists_then="ignore")

    #%% Access the spectra in the database via the get functions.
    # Check what species are in the database at a given temperature (300)
    s300 = db.get(Tgas=300)  # simple use of get
    if verbose:
        print([s.conditions["species"] for s in s300])
    # Check what species are in the database at a given species
    sCO = db.get(species="CO")
    if verbose:
        print([s.conditions["Tgas"] for s in sCO])
    # Advanced use of the get function
    sdb = db.get('Tgas>=600 & Tgas<900 & species=="CO"')
    if verbose:
        print("Number of spectra found for the given conditions: ", len(sdb))

    #%% Do the same in lazy-loading mode, and compare :

    db2 = SpecDatabase(DBNAME, lazy_loading=True)  # create the database

    #%% Access the spectra in the database via the get functions.
    # Check what species are in the database at a given temperature (300)
    s300_2 = db2.get(Tgas=300)  # simple use of get
    print([s.conditions["species"] for s in s300_2])

    assert s300 == s300_2

    # Check what species are in the database at a given species
    sCO_2 = db2.get(species="CO")
    if verbose:
        print([s.conditions["Tgas"] for s in sCO_2])

    assert sCO == sCO_2

    # Advanced use of the get function
    sdb_2 = db2.get('Tgas>=600 & Tgas<900 & species=="CO"')
    if verbose:
        print("Number of spectra found for the given conditions: ", len(sdb))

    assert sdb == sdb_2


if __name__ == "__main__":
    import pytest

    # -s is to plot
    pytest.main(["test_database.py", "-s"])



# -*- coding: utf-8 -*-
"""
Test that RADIS molecule lists are always up-to-date

Note: Radis HITRAN molecules remain hardcoded so we do not require an internet
connection; fetching is only done while running tests
"""

from urllib.request import HTTPError, urlopen

import pytest
from bs4 import BeautifulSoup
from tqdm import tqdm


def fetch_ExoMol_molecules():
    """Get list of molecules in ExoMol by scrapping the ExoMol page"""

    url = "https://exomol.com/data/molecules/"
    try:
        response = urlopen(url).read()
    except HTTPError as err:
        raise ValueError(f"HTTPError opening url={url}") from err

    soup = BeautifulSoup(
        response, features="lxml"
    )  # make soup that is parse-able by bs

    # All others
    rows = soup.find_all(
        "a", {"class": "list-group-item link-list-group-item molecule_link"}
    )
    molecules = [r.get_attribute_list("href")[0] for r in rows]
    molecules = [molecule.split("/")[-2] for molecule in molecules]  # 03/07/2024

    return molecules


@pytest.mark.fast  # this test fails as soon as ExoMol adds a new molecule (often!)
@pytest.mark.needs_connection
def test_ExoMol_molecules_list(verbose=True, *args, **kwargs):
    """Test that ExoMol molecule list in RADIS remains up to date"""

    from radis.db.classes import EXOMOL_MOLECULES
    from radis.misc.basics import compare_lists

    molecules = fetch_ExoMol_molecules()

    if verbose:
        print("ExoMol molecules, fetched online:")
        print(sorted(molecules))

        print("Comparing Radis hardcoded ExoMol molecules list to the ExoMol website:")

    assert (
        compare_lists(
            EXOMOL_MOLECULES,
            molecules,
            l1_str="Radis molecules",
            l2_str="Fetched from ExoMol website",
            print_index=True,
        )
        == 1
    )
    print("----")


def fetch_ExoMol_isotopes(verbose=True, *args, **kwargs):
    """Get dictionary of isotope names in ExoMol by scrapping the ExoMol pages"""

    isotopes_full_names = {}
    molecules = fetch_ExoMol_molecules()

    if verbose:
        print("Parsing ExoMol molecules", flush=True)
    for molecule in tqdm(molecules, disable=not verbose):
        url = f"https://exomol.com/data/molecules/{molecule}"
        try:
            response = urlopen(url).read()
        except HTTPError as err:
            raise ValueError(f"HTTPError opening url={url}") from err

        soup = BeautifulSoup(
            response, features="lxml"
        )  # make soup that is parse-able by bs

        # Recommended database
        rows = soup.find_all("a", {"class": "list-group-item link-list-group-item"})
        isotopes = [r.get_attribute_list("href")[0] for r in rows]

        isotopes_full_names[molecule] = isotopes
    return isotopes_full_names


# %%
def fetch_HITRAN_molecules():
    import re

    url = "https://hitran.org/lbl/"
    try:
        response = urlopen(url).read()
    except HTTPError as err:
        raise ValueError(f"HTTPError opening url={url}") from err

    soup = BeautifulSoup(
        response, features="lxml"
    )  # make soup that is parse-able by bs

    # All others
    rows = soup.find_all("td", {"class": "molecule-formula"})
    molecules = [re.sub(r"<[^>]*>", "", str(r)) for r in rows]

    return molecules


@pytest.mark.fast
@pytest.mark.needs_connection
def test_HITRAN_molecules_list(verbose=True, *args, **kwargs):
    """Test that HITRAN molecule list in RADIS remains up to date"""

    from radis.db.classes import HITRAN_MOLECULES
    from radis.misc.basics import compare_lists

    molecules = fetch_HITRAN_molecules()

    if verbose:
        print("HITRAN molecules, fetched online:")
        print(molecules)

        print("Comparing Radis hardcoded HITRAN molecules list to the HITRAN website:")

    assert (
        compare_lists(
            HITRAN_MOLECULES,
            molecules,
            l1_str="Radis molecules",
            l2_str="Fetched from HITRAN website",
            print_index=True,
        )
        == 1
    )
    print("----")


# %%


def generate_molparam_for_non_HITRAN_species():
    """Simple estimator of molecular parameters when not given in HITRAN.

    isotopic abundance  : in a simple approach, we multiply isotopic abundances
    of all elements of the molecule .  We typically get 0.04% error for main isotope,
    3-10% error in general , up to 80% in the worst cases, so always the same order of magnitude,
    (when compared with species that exist in HITRAN)

    molar mass : we sum the mass of all elements


    Output ``EXTRA_ABUNDANCES_DICT, EXTRA_MOLAR_MASS_DICT, EXTRA_ISOTOPE_FULLNAME_DICT``
    should be pasted in ``default_radis.json``
    """

    # %%
    import numpy as np

    exomol_molecules = fetch_ExoMol_molecules()
    # hitran_molecules = fetch_HITRAN_molecules()
    isotopes_full_names = fetch_ExoMol_isotopes()

    #%%
    import periodictable

    Ia_dict = {}
    Mm_dict = {}
    for atom in [
        "H",
        "C",
        "O",
        "N",
        "S",
        "F",
        "Cl",
        "Br",
        "I",
        "P",
        "Mg",
        "Na",
        "Ni",
        "Al",
        "Cr",
        "Ca",
        "Be",
        "Ti",
        "Fe",
        "Li",
        "Sc",
        "Si",
        "V",
        "Y",
        "K",
        "As",
        "He",
    ]:
        el = getattr(periodictable, atom)
        Ia_dict[atom] = {}
        Mm_dict[atom] = {}
        for iso in el.isotopes:
            if iso.abundance is None:
                Ia_dict[atom][iso.mass_number] = 1e-6  # arbitrary
            else:
                Ia_dict[atom][iso.mass_number] = iso.abundance
            Mm_dict[atom][iso.mass_number] = iso.mass
    # add missing:
    Ia_dict["Al"][26] = 1e-6  # arbitrary
    Mm_dict["Al"][26] = 26.9815385 - 1.006  # doesn't exist by default
    Mm_dict["H"][3] = 2.0141017781 + 1.006  # 3H  'None' by default
    Mm_dict["C"][14] = 13.003354835 + 1.006  # 14C

    import re

    _parse_molecule = re.compile("(\d+)+([A-Z][a-z]?)(\d+)?")

    def get_isotopic_atomic_composition(molecule):
        atoms = {}
        for mass_number, atom, number in _parse_molecule.findall(molecule):
            if number == "":
                number = 1
            if atom in atoms:
                atoms[atom][int(mass_number)] = int(number)
            else:
                atoms[atom] = {int(mass_number): int(number)}
        return atoms

        # out = dict(_parse_molecule.findall(molecule))
        # for k, v in out.items():
        #     if v == '':
        #         out[k] = 1
        #     else:
        #         out[k] = int(v)
        # return out

    # print(get_atoms_in("CO2"))
    print(get_isotopic_atomic_composition("12C-16O2"))

    def estimate_terrestrial_abundance(molecule):
        """``molecule`` given in ExoMol full isotope format"""
        abundance = 1
        atoms_isotopes = get_isotopic_atomic_composition(molecule)
        for atom in atoms_isotopes:
            for Z, N_Z in atoms_isotopes[atom].items():
                abundance *= Ia_dict[atom][Z] ** N_Z
                # if Z in Ia_dict[atom]:
                # else:
                #     abundance *= (1e-6)**N_Z  # arbitrary (low)
        return abundance

    def get_molar_mass(molecule):
        """``molecule`` given in ExoMol full isotope format"""
        atoms_isotopes = get_isotopic_atomic_composition(molecule)
        molar_mass = sum(
            [
                Mm_dict[atom][Z] * N_Z
                for atom in atoms_isotopes
                for Z, N_Z in atoms_isotopes[atom].items()
            ]
        )
        return molar_mass

    assert get_molar_mass("12C-16O2") == 43.98982924

    # Test : compare with terrestrial abundance when known:

    from radis.db.classes import get_molecule
    from radis.db.molparam import MolParams

    molpar = MolParams()

    for _, r in molpar.df.reset_index().iterrows():
        print(get_molecule(r.id), "-", r.isotope_name_exomol)
        if not r.isotope_name_exomol:
            raise
        estimate = estimate_terrestrial_abundance(r.isotope_name_exomol)
        print(
            "... Estim: {0:.3e}".format(
                estimate_terrestrial_abundance(r.isotope_name_exomol)
            )
        )
        print(
            "... Real:  {0:.3e}".format(r.abundance),
            "         ({0:.2%}% error)".format(
                abs(estimate - r.abundance) / r.abundance
            ),
        )

    # %% Generate dictionary of terrestrial abundances for all molecules not in HITRAN
    EXTRA_ABUNDANCES_DICT = {}
    EXTRA_MOLAR_MASS_DICT = {}
    EXTRA_ISOTOPE_FULLNAME_DICT = {}

    print("Check if isotopologues in ExoMol are sorted as in HITRAN")
    for M in exomol_molecules:
        print(M)
        abundances_naive = np.array(
            [
                estimate_terrestrial_abundance(isotope_fullname)
                for isotope_fullname in isotopes_full_names[M]
            ]
        )
        b_sort_naive = np.argsort(abundances_naive)[::-1]  # most abundant isotope first
        abundances_naive = abundances_naive[b_sort_naive]
        molar_mass = [
            get_molar_mass(isotope_fullname)
            for isotope_fullname in np.array(isotopes_full_names[M])[b_sort_naive]
        ]
        # sorted_naive = np.array(isotopes_full_names[M])[b]
        for i, isotope_fullname in enumerate(
            np.array(isotopes_full_names[M])[b_sort_naive]
        ):
            # check if isotopes are sorted by abundances
            try:
                sorted_equally = (
                    molpar.get(M, i + 1, key="isotope_name_exomol") == isotope_fullname
                )
                # sorted_naively = molpar.get(M, i+1, key="isotope_name_exomol") == isotope_fullname
                print(
                    "... ",
                    isotope_fullname,
                    "\t",
                    "ok"
                    if sorted_equally
                    else "⚠️ DIFFERENT ({0})".format(
                        molpar.get(M, i + 1, key="isotope_name_exomol")
                    ),
                )
            except (KeyError, NotImplementedError):
                print("... ", isotope_fullname, "not in HITRAN")
                try:  # add isotope i+1
                    EXTRA_ABUNDANCES_DICT[M][str(i + 1)] = abundances_naive[i]
                    EXTRA_MOLAR_MASS_DICT[M][str(i + 1)] = molar_mass[i]
                    EXTRA_ISOTOPE_FULLNAME_DICT[M][str(i + 1)] = isotope_fullname
                except KeyError:
                    EXTRA_ABUNDANCES_DICT[M] = {str(i + 1): abundances_naive[i]}
                    EXTRA_MOLAR_MASS_DICT[M] = {str(i + 1): molar_mass[i]}
                    EXTRA_ISOTOPE_FULLNAME_DICT[M] = {str(i + 1): isotope_fullname}

    EXTRA_ABUNDANCES_DICT = {
        k: EXTRA_ABUNDANCES_DICT[k] for k in sorted(EXTRA_ABUNDANCES_DICT)
    }
    EXTRA_MOLAR_MASS_DICT = {
        k: EXTRA_MOLAR_MASS_DICT[k] for k in sorted(EXTRA_MOLAR_MASS_DICT)
    }
    EXTRA_ISOTOPE_FULLNAME_DICT = {
        k: EXTRA_ISOTOPE_FULLNAME_DICT[k] for k in sorted(EXTRA_ISOTOPE_FULLNAME_DICT)
    }

    return EXTRA_ABUNDANCES_DICT, EXTRA_MOLAR_MASS_DICT, EXTRA_ISOTOPE_FULLNAME_DICT


# %%

if __name__ == "__main__":
    test_ExoMol_molecules_list()
    # test_HITRAN_molecules_list()
